from zope.interface import alsoProvides, noLongerProvides
from Products.ATContentTypes.content.document import ATDocument
from Products.CMFDiffTool.ATCompoundDiff import ATCompoundDiff

from .BaseTestCase import BaseATTestCase
from zope.component import adapts, provideAdapter
from Products.Archetypes import atapi


class TestATCompoundDiff(BaseATTestCase):
    """Test the portal_diff tool"""

    def testCompoundDiff(self):
        first = self.folder.invokeFactory('Document', 'extended-document')
        first = self.folder[first]
        second = self.folder.invokeFactory('Document', 'extended-document2')
        second = self.folder[second]
        # Change a value
        first.setText('<p>Body1</p>', mimetype='text/html')
        second.setText('<p>Body2</p>', mimetype='text/html')
        fd = ATCompoundDiff(first, second, None)
        # There should be some fields
        self.failUnless(len(fd))
        for field in fd:
            # We've changed the body
            if field.label == 'label_body_text':
                # We have the correct values
                self.assertEqual(field.oldValue, '<p>Body1</p>')
                self.assertEqual(field.newValue, '<p>Body2</p>')
                # And the correctly assigned diff type
                self.assertEqual(field.meta_type, 'HTML Diff')
            # We've also changed the id
            elif field.label == 'label_short_name':
                self.assertEqual(field.oldValue, 'extended-document')
                self.assertEqual(field.newValue, 'extended-document2')
                self.assertEqual(field.meta_type, 'Lines Diff')
            # The dates will be different, but everything else should
            # be the same
            elif field.label not in ('label_creation_date',
                                     'label_modification_date'):
                self.assertEqual(field.oldValue, field.newValue)

    def testAdaptedObjects(self):
        # Add a schema extended field to an ATDocument and test that
        # the compound diff can read it
        from archetypes.schemaextender.field import ExtensionField
        from archetypes.schemaextender.tests.mocks import IHighlighted, Extender

        # custom field class which does not set attributes directly on the
        # content
        class HighlightedField(ExtensionField, atapi.StringField):
            def get(self, instance, **kwargs):
                return IHighlighted.providedBy(instance)

            def getRaw(self, instance, **kwargs):
                return self.get(instance, **kwargs)

            def set(self, instance, value, **kwargs):
                if value and not IHighlighted.providedBy(instance):
                    alsoProvides(instance, IHighlighted)
                elif not value and IHighlighted.providedBy(instance):
                    noLongerProvides(instance, IHighlighted)

        class TestSchemaExtender(Extender):
            adapts(ATDocument)
            fields = [
                HighlightedField('schemaextender_test',
                                 schemata='settings',
                                 widget=atapi.BooleanWidget(
                                                    label="Extended",
                                                    description=""),
                                 ),
                      ]

        """Ensure that tool instances implement the portal_diff interface"""
        provideAdapter(TestSchemaExtender,
                       name=u"archetypes.schemaextender.tests")

        first = self.folder.invokeFactory('Document', 'extended-document')
        first = self.folder[first]
        second = self.folder.invokeFactory('Document', 'extended-document2')
        second = self.folder[second]
        # Change the value
        alsoProvides(second, IHighlighted)
        fd = ATCompoundDiff(first, second, None)
        field = fd[-1]
        self.assertEqual(field.oldValue, False)
        self.assertEqual(field.newValue, True)
        self.assertEqual(field.label, "Extended")



from zope.schema import (Bytes, Iterable, Container, Text, Date, Datetime,
                         Time, Choice, Bool)
from zope.globalrequest import getRequest
from z3c.form.interfaces import INPUT_MODE

from plone.app.textfield import RichText
from plone.autoform.base import AutoFields

from Products.CMFDiffTool.TextDiff import AsTextDiff
from Products.CMFDiffTool.namedfile import NamedFileBinaryDiff
from Products.CMFDiffTool.namedfile import FILE_FIELD_TYPES
from Products.CMFDiffTool.namedfile import NamedFileListDiff
from Products.CMFDiffTool.choicediff import ChoiceDiff
from Products.CMFDiffTool.choicediff import get_schemas
from Products.CMFDiffTool.CMFDTHtmlDiff import CMFDTHtmlDiff
from Products.CMFDiffTool.FieldDiff import FieldDiff
from Products.CMFDiffTool.ListDiff import ListDiff
from Products.CMFDiffTool.TextDiff import TextDiff

# TODO: Perhaps this can be replaced with some kind of Zope 3 style
# adaptation, in order to provide better extensibility.
FIELDS_AND_DIFF_TYPES_RELATION = [
    (FILE_FIELD_TYPES, NamedFileBinaryDiff),
    ((Iterable, Container), ListDiff),
    ((Date, Datetime, Time), AsTextDiff),
    ((Bool, ), AsTextDiff),
    ((Choice, ), ChoiceDiff),
    ((Text, Bytes), TextDiff),
    ((RichText, ), CMFDTHtmlDiff),
]

"""
Relates field types (`zope.schema.Field` subclasses) and "diff types"
(`Products.CMFEditions.BaseDiff.BaseDiff` subclasses).

To find the best diff type for a field type this list will be searched until
a match is found.
If a match is not found then `FALL_BACK_DIFF_TYPE` is used.
"""

FALL_BACK_DIFF_TYPE = FieldDiff

# TODO: Perhaps this is not the best approach. Instead we should write a diff
# type which can handle lists of any kind of elements, providing extensibility
# mechanisms if we want to specialize the handling of certain value types.
# (rafaelbco)
VALUE_TYPES_AND_DIFF_TYPES_RELATION = [
    (FILE_FIELD_TYPES, NamedFileListDiff),
]

"""
When a field is detected to be a list-like field we use this list in the same
fashion as `FIELDS_AND_DIFF_TYPES_RELATION` to try to find the best "diff type
" according to the "value type" of the field, i.e the type of the elements in
the list. If a match is not found then a fall back is used.
"""

# TODO: provide an easier way to exclude fields.
EXCLUDED_FIELDS = ('modification_date', 'changeNote')
"""Names of fields not to compare."""


class DexterityCompoundDiff(object):
    """Same as `Products.CMFDiffTool.ATCompoundDiff.ATCompoundDiff` but for
    Dexterity.
    """

    meta_type = 'Compound Diff for Dexterity types'

    def __init__(self, obj1, obj2, field, id1=None, id2=None):
        self.id1 = id1 or obj1.getId()
        self.id2 = id2 or obj2.getId()
        self.obj1 = obj1
        self._diffs = self._diff(obj1, obj2)

    def __getitem__(self, index):
        return self._diffs[index]

    def __len__(self):
        return len(self._diffs)

    def __iter__(self):
        return iter(self._diffs)

    def _diff(self, obj1, obj2):
        """
        Compute the differences between 2 objects.

        Return: a sequence of `IDifference` objects.
        """
        (default_schema, additional_schemata) = get_schemas(obj1)
        diffs = self._diff_schema(obj1, obj2, default_schema, 'default')

        for schema in additional_schemata:
            diffs.extend(self._diff_schema(obj1, obj2, schema, 'metadata'))

        return diffs

    def _diff_schema(self, obj1, obj2, schema, schema_name):
        """
        Compute the differences between 2 objects in respect to the given
        schema interface.

        Return: a sequence of `IDifference` objects.
        """
        return [
            self._diff_field(obj1, obj2, schema[name], schema_name)
            for name in self._compute_fields_order(schema)
            if name not in EXCLUDED_FIELDS
        ]

    def _diff_field(self, obj1, obj2, field, schema_name):
        """
        Compute the differences between 2 objects in respect to the given
        field (`zope.schema.Field` instance).

        Return: an `IDifference` object.
        """
        diff_type = self._get_diff_type(field)
        return diff_type(
            obj1,
            obj2,
            field.getName(),
            id1=self.id1,
            id2=self.id2,
            field_name=field.getName(),
            field_label=field.title,
            schemata=schema_name
        )

    def _get_diff_type(self, field):
        """
        Return a subclass of `Products.CMFEditions.BaseDiff.BaseDiff` suitable
        for the given `zope.schema.Field` instance.
        """
        diff_type = self._compute_diff_type(
            field, FIELDS_AND_DIFF_TYPES_RELATION)

        if diff_type is ListDiff:
            return (
                self._compute_diff_type(
                    field.value_type, VALUE_TYPES_AND_DIFF_TYPES_RELATION)
                or diff_type
            )

        return diff_type or FALL_BACK_DIFF_TYPE

    def _compute_diff_type(self, field, relation):
        """
        Return the best "diff type" (subclass of
        `Products.CMFEditions.BaseDiff.BaseDiff`) suitable for the given
        `zope.schema.Field` instance according to `relation`. The `relation`
        is searched until a match is found. Return `None` otherwise.

        Parameters:
        field -- `zope.schema.Field` instance.
        relation -- Sequence of tuples (field_types, diff_type) where
            field_types is a tuple of `zope.schema.Field` subclasses
            and diff_type is a `Products.CMFEditions.BaseDiff.BaseDiff`
            subclass.
        """

        for (field_types, diff_type) in relation:
            if isinstance(field, field_types):
                return diff_type

        return None

    def _compute_fields_order(self, schema):
        """
        Given a `schema` interface compute the field ordering the way
        `plone.autoform` does, i.e taking into account `plone.directives.form`
        ordering directives.

        Return: a list of field names in order.
        """
        auto_fields = AutoFields()
        auto_fields.context = self.obj1 # See issue #330
        auto_fields.schema = schema
        auto_fields.request = getRequest()
        auto_fields.mode = INPUT_MODE
        auto_fields.updateFieldsFromSchemata()
        return auto_fields.fields



# -*- coding: utf-8 -*-
"""Initialize CMFDiffTool Product"""
# Set up a MessageFactory for the cmfdifftool domain
from zope.i18nmessageid import MessageFactory
CMFDiffToolMessageFactory = MessageFactory('cmfdifftool')

from Products.CMFCore.utils import ToolInit

from Products.CMFDiffTool import CMFDiffTool
from Products.CMFDiffTool import FieldDiff
from Products.CMFDiffTool import TextDiff
from Products.CMFDiffTool import ListDiff
from Products.CMFDiffTool import BinaryDiff
from Products.CMFDiffTool import CMFDTHtmlDiff
from Products.CMFDiffTool import ATCompoundDiff

CMFDiffTool.registerDiffType(BinaryDiff.BinaryDiff)
CMFDiffTool.registerDiffType(FieldDiff.FieldDiff)
CMFDiffTool.registerDiffType(ListDiff.ListDiff)
CMFDiffTool.registerDiffType(TextDiff.TextDiff)
CMFDiffTool.registerDiffType(TextDiff.AsTextDiff)
CMFDiffTool.registerDiffType(CMFDTHtmlDiff.CMFDTHtmlDiff)
CMFDiffTool.registerDiffType(ATCompoundDiff.ATCompoundDiff)

# Soft plone.namedfile dependency
try:
    from Products.CMFDiffTool import namedfile
except ImportError:
    pass
else:
    CMFDiffTool.registerDiffType(namedfile.NamedFileBinaryDiff)
    CMFDiffTool.registerDiffType(namedfile.NamedFileListDiff)

# Soft Dexterity dependency
try:
    from Products.CMFDiffTool import dexteritydiff
    from Products.CMFDiffTool import choicediff
except ImportError:
    pass
else:
    CMFDiffTool.registerDiffType(choicediff.ChoiceDiff)
    CMFDiffTool.registerDiffType(dexteritydiff.DexterityCompoundDiff)

tools = (CMFDiffTool.CMFDiffTool, )


def initialize(context):
    ToolInit('CMF Diff Tool',
             tools=tools,
             icon='tool.gif'
             ).initialize(context)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

version = '3.0.1'

setup(name='Products.CMFDiffTool',
      version=version,
      description="Diff tool for Plone",
      long_description=open("README.txt").read() + "\n" +
                       open("CHANGES.txt").read(),
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 2.7",
        "Framework :: Plone",
        "Framework :: Plone :: 5.0",
        "Topic :: Internet :: WWW/HTTP",
        ],
      keywords='Diff Plone',
      author='Brent Hendricks',
      author_email='plone-developers@lists.sourceforge.net',
      url='http://pypi.python.org/pypi/Products.CMFDiffTool',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      extras_require=dict(
        test=[
            'zope.component',
            'plone.app.testing',
            'plone.namedfile',
            'plone.app.dexterity',
            'plone.app.contenttypes',
        ]
      ),
      install_requires=[
        'setuptools',
        'zope.interface',
        'Products.CMFCore',
        'Products.GenericSetup',
        'Acquisition',
        'Zope2',
      ],
      entry_points="""
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



# -*- coding: utf-8 -*-
#
# CMFDiffTool tests
#
from os import linesep
from Testing import ZopeTestCase
from Products.CMFDiffTool.FieldDiff import FieldDiff

_marker = []


class A:
    attribute = "value"

    def method(self):
        return "method value"


class B:
    attribute = "different value"

    def method(self):
        return "different method value"


class TestFieldDiff(ZopeTestCase.ZopeTestCase):
    """Test the FieldDiff class"""

    def testInterface(self):
        """Ensure that tool instances implement the portal_diff interface"""
        from Products.CMFDiffTool.interfaces.portal_diff import IDifference
        self.failUnless(IDifference.implementedBy(FieldDiff))

    def testAttributeSame(self):
        """Test attribute with same value"""
        a = A()
        fd = FieldDiff(a, a, 'attribute')
        self.failUnless(fd.same)

    def testMethodSame(self):
        """Test method with same value"""
        a = A()
        fd = FieldDiff(a, a, 'method')
        self.failUnless(fd.same)

    def testAttributeDiff(self):
        """Test attribute with different value"""
        a = A()
        b = B()
        fd = FieldDiff(a, b, 'attribute')
        self.failIf(fd.same)

    def testMethodDiff(self):
        """Test method with different value"""
        a = A()
        b = B()
        fd = FieldDiff(a, b, 'method')
        self.failIf(fd.same)

    def testGetLineDiffsSame(self):
        """test getLineDiffs() method with same value"""
        a = A()
        fd = FieldDiff(a, a, 'attribute')
        expected = [('equal', 0, 1, 0, 1)]
        self.assertEqual(fd.getLineDiffs(), expected)

    def testGetLineDiffsDifferent(self):
        """test getLineDiffs() method with different value"""
        a = A()
        b = B()
        fd = FieldDiff(a, b, 'attribute')
        expected = [('replace', 0, 1, 0, 1)]
        self.assertEqual(fd.getLineDiffs(), expected)

    def testSameText(self):
        """Test text diff output with same value"""
        a = A()
        fd = FieldDiff(a, a, 'attribute')
        self.assertEqual(fd.ndiff(), '  value')

    def testDiffText(self):
        """Test text diff output with different value"""
        a = A()
        b = B()
        expected = "- value%s+ different value" % linesep
        fd = FieldDiff(a, b, 'attribute')
        self.assertEqual(fd.ndiff(), expected)
